/*
 * tmos- printk.c
 *
 * Copyright (C) 2009  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/serial.h>
#include <stdarg.h>

#define putc(x) ser_txc(x)

#define SIGN	2		/* unsigned/signed long */
#define HEX	4		/* print hex number */
#define ZEROS	8		/* print heading zeros */

static char d_to_char(char x) __preempt
{
	return ((x >= 10) ? 'a' - 10 : '0') + x;
}

static void print_int(unsigned int val, unsigned char width,
		unsigned char flags) REENTRANT
{
	unsigned char x;

	if (flags & ZEROS) {
		switch (width) {
		case 1: goto print_1;
		case 2: goto print_10;
		case 3: goto print_100;
		case 4: goto print_1000;
		default: goto print_10000;
		}
	}
	if (val < 10)
		goto print_1;
	if (val < 100)
		goto print_10;
	if (val < 1000)
		goto print_100;
	if (val < 10000)
		goto print_1000;
	
print_10000:
	x = val / 10000;
	putc(d_to_char(x));
	val -= 10000 * x;
	
print_1000:
	x = val / 1000;
	putc(d_to_char(x));
	val -= 1000 * x;
	
print_100:
	x = val / 100;
	putc(d_to_char(x));
	val -= 100 * x;

print_10:
	x = val / 10;
	putc(d_to_char(x));
	val -= 10 * x;
print_1:
	x = val;
	putc(d_to_char(x));
}

static void print_hex(unsigned int val) REENTRANT
{
	if (val < 0x10)
		goto print_1;
	if (val < 0x100)
		goto print_10;
	if (val < 0x1000)
		goto print_100;
	putc(d_to_char((val >> 12) & 0xf));
print_100:
	putc(d_to_char((val >> 8) & 0xf));
print_10:
	putc(d_to_char((val >> 4) & 0xf));
print_1:
	putc(d_to_char(val & 0xf));
}

/*
 * printk - print a message
 * @fmt - format string
 *
 * This function is a reentrant printf equivalent which may be called
 * from any context.
 *
 * See also:
 * printf(3)
 */
int printk(const char *fmt, ...) REENTRANT
{
	va_list args;
	int ret = 0;
	unsigned int flags = 0;
	unsigned long num;
	unsigned char width = 0;

	va_start(args, fmt);
	for (; *fmt; fmt++) {
		unsigned char c = *fmt;
		
		if (c != (unsigned char) '%') {
			ser_txc(c);
			continue;
		}
		flags = 0;
		fmt++;
		if (*fmt == '0') {
			flags |= ZEROS;
			fmt++;
		}
		if (*fmt >= '1' && *fmt <= '9') {
			width = *fmt - '0';
			fmt++;
		}
		if (*fmt == 's') {
			char *s = va_arg(args, char *);
			for (; *s; s++)
				ser_txc(*s);
			continue;
		}
		switch (*fmt) {
		case 'u':
			break;
		case 'd':
			flags |= SIGN;
			break;
		case 'x':
			flags |= HEX;
			break;
		}
		
		num = va_arg(args, unsigned int);
		if (flags & SIGN) {
			num = (signed int) num;
		}
		if (flags & SIGN) {
			if ((signed long) num < 0) {
				putc('-');
				num = - (signed long) num;
			}
		}
		if (flags & HEX) {
			print_hex(num);
		} else {
			print_int(num, width, flags);
		}
	}
	va_end(args);

	return ret;
}
